let software_name = "check_caml_version";;
let software_version = "0.1.0";;

let print_version () =
  Format.printf "%s\n" software_version;
  exit 0
;;

let get_usage () =
  Printf.sprintf
    "Usage: %s [options] [<caml_version> <comparison> <caml_template>]\
   \n  <caml_version> is a valid Caml system version\
   \n  <comparison> is one of: -eq, -ne, -lt, -gt, -le, -ge\
   \n  <caml_template> is a valid Caml system version, possibly truncated\
   \n\
   \n  options are:\
   \n  -version: prints the %s version"
    software_name
    software_name
;;

let print_help () =
  print_endline (get_usage ());
  exit 0
;;

type comparison =
  | Eq
  | Ne
  | Le
  | Ge
  | Lt
  | Gt
;;

let string_of_comparison = function
  | Eq -> "Eq"
  | Ne -> "Ne"
  | Le -> "Le"
  | Ge -> "Ge"
  | Lt -> "Lt"
  | Gt -> "Gt"
;;

let set_caml_version, get_caml_version =
  let caml_version = ref None in
  (fun s -> caml_version := Some s),
  (fun () ->
   match !caml_version with
   | Some version -> version
   | None -> failwith "No Caml version")
;;

let set_compare_version, get_compare_version =
  let compare_version = ref false in
  (fun () -> compare_version := true),
  (fun () -> !compare_version)
;;

let set_compare_major_version, get_compare_major_version =
  let compare_major_version = ref false in
  (fun () -> compare_major_version := true),
  (fun () -> !compare_major_version)
;;

let set_compare_minor_version, get_compare_minor_version =
  let compare_minor_version = ref false in
  (fun () -> compare_minor_version := true),
  (fun () -> !compare_minor_version)
;;

let set_compare_patch_version, get_compare_patch_version =
  let compare_patch_version = ref false in
  (fun () -> compare_patch_version := true),
  (fun () -> !compare_patch_version)
;;

let set_compare_template, get_compare_template =
  let compare_template = ref None in
  (fun s -> compare_template := Some s),
  (fun () ->
   match !compare_template with
   | Some template -> template
   | None -> failwith "No version template")
;;

let set_comparison, get_comparison =
  let comparison = ref None in
  (fun comp -> comparison := Some comp),
  (fun () ->
   match !comparison with
   | Some comp -> comp
   | None -> failwith "No comparison set")
;;

let set_compare_with_eq s = set_comparison Eq; set_compare_template s;;

let set_compare_with_ne s = set_comparison Ne; set_compare_template s;;

let set_compare_with_le s = set_comparison Le; set_compare_template s;;

let set_compare_with_lt s = set_comparison Lt; set_compare_template s;;

let set_compare_with_ge s = set_comparison Ge; set_compare_template s;;

let set_compare_with_gt s = set_comparison Gt; set_compare_template s;;

let parse_args argv =
  Arg.parse_argv argv [
    ("-full-version", Arg.Unit set_compare_version,
     "treat complete version");
    ("-major-version", Arg.Unit set_compare_major_version,
     "treat major version number");
    ("-minor-version", Arg.Unit set_compare_minor_version,
     "treat minor version number");
    ("-patch-version", Arg.Unit set_compare_patch_version,
     "treat patch version number");
    ("-eq", Arg.String set_compare_with_eq,
     "compare using equal");
    ("-ne", Arg.String set_compare_with_ne,
     "compare using not equal");
    ("-le", Arg.String set_compare_with_le,
     "compare using lesser or equal");
    ("-lt", Arg.String set_compare_with_lt,
     "compare using lesser than");
    ("-ge", Arg.String set_compare_with_ge,
     "compare using gretter or equal");
    ("-gt", Arg.String set_compare_with_gt,
     "compare using greater than");
    ("-v", Arg.Unit print_version,
     ": print the version of check_caml_version");
    ("-h", Arg.Unit print_help,
     ": print some help about check_caml_version");
    ("-help", Arg.Unit print_help,
     ": print some help about check_caml_version");
  ]
  set_caml_version
  (get_usage ())
;;

let scan_version version =
  let ib = Scanf.Scanning.from_string version in

  let add_part l part =
    let part_option =
      if part <> "" then Some part else None in
    part_option :: l in

  let scan_major l =
    Scanf.bscanf ib "%_[^0-9]%[0-9]" (add_part l) in
  let scan_minor l =
    if Scanf.Scanning.end_of_input ib then add_part l "" else
    Scanf.bscanf ib ".%[0-9]" (add_part l) in
  let scan_patch = scan_minor in

  let parts = scan_patch (scan_minor (scan_major [])) in

  List.rev parts
;;

let rec eq_parts template_parts version_parts =
  match template_parts, version_parts with
  | [], [] -> true
  | None :: template_parts, None :: version_parts ->
    eq_parts template_parts version_parts
  | Some major_template :: template_parts,
    Some major_version :: version_parts ->
    major_template = major_version &&
    eq_parts template_parts version_parts
  | _ -> false
;;

let rec gt_parts template_parts version_parts =
  match template_parts, version_parts with
  | [], [] -> false
  | None :: template_parts, None :: version_parts ->
    gt_parts template_parts version_parts
  | Some major_template :: template_parts,
    Some major_version :: version_parts ->
    major_template > major_version ||
    major_template = major_version &&
    gt_parts template_parts version_parts
  | Some major_template :: _, None :: _ ->
    (* None is considered as zero: for instance 4 is 4.0 *)
    int_of_string major_template > 0
  | _ -> false
;;

let rec lt_parts template_parts version_parts =
  match template_parts, version_parts with
  | [], [] -> false
  | None :: template_parts, None :: version_parts ->
    lt_parts template_parts version_parts
  | Some major_template :: template_parts,
    Some major_version :: version_parts ->
    major_template < major_version ||
    major_template = major_version &&
    lt_parts template_parts version_parts
  | None :: _, Some major_version :: _ ->
    (* None is considered as zero: for instance 4 is 4.0 *)
    0 < int_of_string major_version
  | _ -> false
;;

let check_version version comparison template =
  let template_parts = scan_version template in
  let version_parts = scan_version version in
  match comparison with
  | Eq -> eq_parts version_parts template_parts
  | Ne -> not (eq_parts version_parts template_parts)
  | Gt -> gt_parts version_parts template_parts
  | Le -> not (gt_parts version_parts template_parts)
  | Lt -> lt_parts version_parts template_parts
  | Ge -> not (lt_parts version_parts template_parts)
;;

let main () =
  parse_args Sys.argv;
  let version = get_caml_version ()
  and template = get_compare_template ()
  and comparison = get_comparison () in
(*
  prerr_endline (Printf.sprintf
    "version = %s; template = %s; comparison = %s"
    version template (string_of_comparison comparison));
*)
  let result = check_version version comparison template in
(*
  prerr_endline (Printf.sprintf
    "result is %B"
    result);
*)
  match result with
  | true -> print_string "1\n"
  | false -> print_string "0\n"
;;

main ();;

(*
 Local Variables:
  compile-command: "ocamlc -o check_caml_version check_caml_version.ml"
  End:
*)
