(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ufbuffer.ml,v 1.7 2017/05/31 15:24:06 fclement Exp $ *)

(* Flushing buffers. *)

module Nat = Unatural;;
module B = Ubytes;;

let _0_ = Nat._0_
and _1_ = Nat._1_
;;

type length = B.length
and index = B.index
;;

type t =
  {
   buffer : B.t;
   mutable position : index;
   length : length;
   flush : string -> unit;
  }
and fbuffer = t
;;

let create f n =
  let n = max _1_ n in
  let s = B.create n in
  {
   buffer = s;
   position = _0_;
   length = n;
   flush = f;
  }
;;

let contents b = B.sub_string b.buffer _0_ b.position;;

let flush b =
  if b.position > _0_ then begin
    b.flush (contents b);
    b.position <- _0_;
  end
;;

let add_char b c =
  if b.position >= b.length then flush b;
  (* Now, we have 0 <= b.position < b.length (see flush).
     Therefore, the following B.unsafe_set will be safe. *)
  let pos = b.position in
  B.unsafe_set b.buffer pos c;
  b.position <- Nat.succ pos;
;;

let add_bytes b s =
  let blen = b.length
  and slen = B.length s in
  let rec loop i len =
    (* i is increasing from 0 and len is decreasing from slen. *)
    if len > _0_ then
    let pos = b.position in
    (* We always have pos <= blen.
       Therefore, the following Nat.sub will not fail. *)
    let available = Nat.sub blen pos in
    if available > _0_ then begin
      let added = min len available in
      (* now, we have added <= len.
         Therefore, the following Nat.sub will not fail. *)
      B.blit s i b.buffer pos added;
      b.position <- Nat.add pos added;
      loop (Nat.add i added) (Nat.sub len added)
    end else begin
      flush b;
      loop i len
    end in
  loop _0_ slen
;;

let add_string b s = add_bytes b (B.unsafe_of_string s);;
