(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: umatrix_types.ml,v 1.16 2017/05/31 15:24:06 fclement Exp $ *)

(* The private type of matrices. *)

module Nat = Unatural;;
module A = Uarray;;

let _0_ = Nat._0_
and _1_ = Nat._1_
;;

let qual = Upervasives.qualify "Umatrix";;
let invalid_arg = Upervasives.invalid_arg_qual qual
;;

type 'a t = 'a array array
and 'a matrix = 'a t
;;

type length = A.length
and index = A.index

and line_number = length
and column_number = length
and line_index = index
;;

type size = line_number * column_number;;

let empty = [||];;

let gen_msg i len_aai nj =
  Printf.sprintf
    "non rectangular argument, line %d is of length %d\
     but was expected of length %d"
    (Nat.to_int i) (Nat.to_int len_aai) (Nat.to_int nj)
;;

let get_array_array_line_length aa =
  A.get_same qual "of_array_array" gen_msg A.length aa
;;

let to_t aa =
  let ni = A.length aa in
  if ni = _0_ then empty else
  let nj = get_array_array_line_length aa in
  if nj = _0_ then empty else
  aa
;;

external of_t : 'a t -> 'a array array = "%identity";;

let dim_i m = A.length (of_t m);;

let size m =
  let ni = dim_i m in
  if ni = _0_ then (_0_, _0_) else
  (* Now, we have ni > 0.
     Therefore, the following A.unsafe_get will be safe. *)
  let aa = of_t m in
  let nj = A.length (A.unsafe_get aa _0_) in
  (ni, nj)
;;

let dim_j m = snd (size m);;

let check_bound msg fun_name n len =
  if n >= len then
    invalid_arg fun_name
      (Printf.sprintf "%s index %d out of bounds" msg (Nat.to_int n))
;;

let check_line_bound = check_bound "line";;

let check_array_length fun_name a len =
  if A.length a <> len then invalid_arg fun_name "wrong array length"
;;

let unsafe_set_line m = A.unsafe_set (of_t m);;

let set_line m i ai =
  let (ni, nj) = size m in
  check_line_bound "set_line" i ni;
  check_array_length "set_line" ai nj;
  (* Now, we have 0 <= i < ni and length ai = nj.
     Therefore, the following unsafe_set_line will be safe. *)
  unsafe_set_line m i ai
;;

(* We could add here features fill_lines and blit_lines... *)
