(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ustring.mli,v 1.42 2017/05/31 15:24:06 fclement Exp $ *)

(** {3 String operations}

 Length, index and offset arguments are of type [Unatural.t]. *)

type t = string
(** Alias for type [string]. *)
;;

val empty : t
(** A string of length 0. *)
;;

type length = Upervasives.length
and index = Upervasives.index
and offset = Upervasives.offset
(** The types for lengths, indices, and offsets;
 these are aliases for type Unatural.t. *)
;;

val length : t -> length
(** Return the length (number of characters) of the given string. *)
;;

val get : t -> index -> char
(** [get s n] returns the character at index [n] in string [s].

 Raise [Invalid_argument] if [n] is not a valid index in [s].*)
;;

val of_char : char -> t
(** [of_char c] returns a new string of length [1] that contains character
 [c]. *)
;;

val make : length -> char -> t
(** [make n c] returns a new string of length [n], filled with character
 [c].

 Raise [Invalid_argument] if [n > Sys.max_string_length]. *)
;;

val init : length -> (index -> char) -> t
(** [init n f] returns a new string of length [n], with character number [i]
 initialized to the result of [f i].
 In other terms, [init n f] tabulates the results of [f] applied to the
 integers [0..n-1] (in increasing order).

 Raise [Invalid_argument] if [n > Sys.max_string_length]. *)
;;

val sub : t -> offset -> length -> t
(** [sub s ofs len] returns a new string of length [len] that contains the
 substring of [s] that starts at index [ofs] and has length [len].

 Raise [Invalid_argument] if [ofs] and [len] do not designate a valid
 substring of [s]. *)
;;

val blit : t -> offset -> Ubytes.t -> offset -> length -> unit
(** [blit src srcofs dst dstofs len] modifies byte sequence [dst] in place,
 copying [len] characters from string [src], starting at index [srcofs], to
 byte sequence [dst], starting at index [dstofs].

 Raise [Invalid_argument] if [srcofs] and [len] do not designate a valid
 substring of [src], or if [dstofs] and [len] do not designate a valid
 substring of [dst]. *)
;;

val concat : t -> t list -> t
(** [concat sep sl] returns a new string that contains the concatenation of
 strings of the list [sl], inserting the separator string [sep] between each.

 Raise [Invalid_argument] if the result would be longer than
 [Sys.max_string_length] characters. *)
;;

val cat : t -> t -> t
(** [cat s1 s2] returns a new string that contains the concatenation of
 strings [s1] and [s2] (in that order).

 Raise [Invalid_argument] if the result would be longer than
 [Sys.max_string_length] bytes. *)
;;

val replace : t -> offset -> length -> t -> offset -> length -> t
(** [replace s1 ofs1 len1 s2 ofs2 len2] returns a new string that contains a
 copy of [s1] with the substring starting at index [ofs1] and of length
 [len1] replaced by the substring of [s2] starting at index [ofs2] and of
 length [len2].
 Lengths [len1] and [len2] may differ.

 Raise [Invalid_argument] if [ofs1] and [len1] do not designate a valid
 substring of [s1], or if [ofs2] and [len2] do not designate a valid
 substring of [s2]. *)
;;

val iter : (char -> unit) -> t -> unit
(** [iter f s] applies function [f] in turn to all the characters of string
 [s].
 It is equivalent to
 [f (get s] {!Unatural._0_}[); f (get s] {!Unatural._1_}[); ...;
 f (get s (]{!Unatural.pred} [(]{!length} [s)); ()]. *)
;;

val iter_i : (index -> char -> unit) -> t -> unit
(** Same as [iter], but the function is applied to the index of the
 character as first argument and the character itself as second argument. *)
;;

val map : (char -> char) -> t -> t
(** [map f s] returns a new string that contains the application of function
 [f] in turn to all the characters of string [s] (in increasing index
 order). *)
;;

val map_i : (index -> char -> char) -> t -> t
(** Same as [map], but the function is applied to the index of the character
 as first argument and the character itself as second argument. *)
;;

val fold_left : ('accu -> char -> 'accu) -> 'accu -> t -> 'accu
(** [fold_left f a s] is
 [f (... (f (f a s.[0]) s.[1]) ...) s.[]{!length} [s - 1]]. *)
;;

val trim : t -> t
(** Return a new string that contains a copy of the argument, without leading
 and trailing whitespace.
 The characters regarded as whitespace are:
 [' ', '\012', '\n', '\r', and '\t'].
 If there is neither leading nor trailing whitespace character in the
 argument, return the original string itself, not a copy. *)
;;

val trim_edges : (char -> bool) -> t -> t
(** [trim_edges p s] returns a new string that contains a copy of string [s]
 without leading and trailing characters that satisfy predicate [p].
 It returns [s] itself, not a copy of it, when there is no such leading nor
 trailing characters in [s]. *)
;;

val trim_consecutive : (char -> bool) -> t -> t
(** [trim_consecutive p s] returns a new string that contains a copy of
 string [s] with no repetition of characters that satisfy predicate [p].
 It returns [s] itself, not a copy of it, when there is no such repetition in
 [s]. *)
;;

val trim_doubles : t -> t
(** [trim_doubles s] returns a new string that contains a copy of string [s]
 with no repeated characters.
 It returns [s] itself, not a copy of it, when there is no repetition in
 [s]. *)
;;

val escaped : t -> t
(** Return a new string that contains a copy of the string argument, with
 special characters represented by escape sequences, following the lexical
 conventions of OCaml.
 All characters outside the ASCII printable range (32..126) are escaped,
 as well as backslash and double-quote.
 Return the original string itself, not a copy of iy, when there is no
 special character in it that needs escaping.

 Raise [Invalid_argument] if the result would be longer than
 [Sys.max_string_length] bytes.

 The function [Scanf.unescaped] is a left inverse of [escaped], i.e.
 [Scanf.unescaped (escaped s) = s] for any string [s] (unless [escape s]
 fails). *)
;;

type escape_char = char
and escaped_chars = char list
(** The types of escaping characters and of characters to escape. *)
;;

val escape_chars : escape_char -> escaped_chars -> t -> t
(** [escape_chars esc cl s] returns a new string that contains a copy of
 string [s] where all occurrences of characters from list [cl] are escaped
 with character [esc].
 If [esc] does not belong to [cl], it is automatically added.
 It is harmless, but useless, to duplicate characters in list c[l].

 Raise [Invalid_argument] if the result would be longer than
 [Sys.max_string_length]. *)
;;

val unescape_chars : escape_char -> escaped_chars -> t -> t
(** [unescape_chars esc cl s] returns a new string that contains a copy of
 string [s] where the escape character [esc] is removed from all escaped
 sequences (character [esc] followed by a character from list [cl]).
 If [esc] does not belong to [cl], it is automatically added.
 It is harmless, but useless, to duplicate characters in list [cl].

 Raise [Failure] if string [s] contains an occurrence of [esc] not followed
 by a character from [cl]. *)
;;

val index : t -> char -> index
(** [index s c] returns the index of the first occurrence of character [c] in
 string [s].

 Raise [Not_found] if [c] does not occur in [s]. *)
;;

val rindex : t -> char -> index
(** [rindex s c] returns the index of the last occurrence of character [c] in
 string [s].

 Raise [Not_found] if [c] does not occur in [s]. *)
;;

val index_from : t -> index -> char -> index
(** [index_from s i c] returns the index of the first occurrence of character
 [c] in string [s] starting from index [i] (included).

 Raise [Invalid_argument] if [i] is not a valid index in [s].
 Raise [Not_found] if [c] does not occur in [s] from position [i]. *)
;;

val rindex_from : t -> index -> char -> index
(** [rindex_from s i c] returns the index of the last occurrence of character
 [c] in string [s] up to index [i] (included).

 Raise [Invalid_argument] if [i] is not a valid index in [s].
 Raise [Not_found] if [c] does not occur in [s] before position [i]. *)
;;

val indices : t -> char -> index list
(** [indices s c] returns the list of indices of al the occurrences of
 character [c] in string [s]. *)
;;

val contains : t -> char -> bool
(** [contains s c] tests if character [c] appears in string [s]. *)
;;

val contains_from : t -> index -> char -> bool
(** [contains_from s i c] tests if character [c] appears in string [s]
 starting from index [i] (included).

 Raise [Invalid_argument] if [i] is not a valid index in [s]. *)
;;

val rcontains_from : t -> index -> char -> bool
(** [rcontains_from s i c] tests if character [c] appears in string [s]
 up to index [i] (included).

 Raise [Invalid_argument] if [i] is not a valid index in [s]. *)
;;

val gather : char list -> t
(** [gather l] returns a new string that contains the concatenation of
 characters of list [l], in the same order.

 Raise [Invalid_argument] if the character list [l] is longer than
 [Sys.max_string_length]. *)
;;

val scatter : t -> char list
(** [scatter s] returns the list of characters of string [s], in the same
 order. *)
;;

val uppercase : t -> t
(** Return a new string that contains a copy of the string argument, with all
 lowercase letters translated to uppercase, using the US ASCII character
 set. *)
;;

val lowercase : t -> t
(** Return a new string that contains a copy of the string argument, with all
 uppercase letters translated to lowercase, using the US ASCII character
 set. *)
;;

val capitalize : t -> t
(** Return a new string that contains a copy of the string argument, with the
 first character translated to uppercase, using the US ASCII character
 set. *)
;;

val uncapitalize : t -> t
(** Return a new string that contains a copy of the string argument, with the
 first character translated to lowercase, using the US ASCII character
 set. *)
;;

val glue_line : escape_char -> char -> t list -> t
(** [glue_line esc sep sl] is similar to {!concat}, except that separator
 [sep] is a single character that is escaped using character [esc] in each
 item of the string list [sl] before concatenation occurs.

 Applying {!explode_line} [esc sep] to the result gives [sl] back.

 Raise [Invalid_argument] if the result would be longer than
 [Sys.max_string_length]. *)
;;

val explode_line : escape_char -> char -> t -> t list
(** [explode_line esc sep s] uses separator character [sep] to explode string
 [s] into a string list where escaped occurrences of character [sep] are
 unescaped.

 Applying {!glue_line} [esc sep] to the result gives [s] back.

 Raise [Failure] if string [s] contains an occurrence of [esc] neither
 followed by [sep], nor [esc]. *)
;;

type prefix = t
and suffix = t
(** The types for prefixes and suffixes;
 these are aliases for type [string]. *)
;;

val check_prefix : prefix -> t -> bool
(** [check_prefix pref s] checks that string [s] starts with prefix [pref].
 Returns [true] when [pref] is empty. *)
;;

val check_suffix : suffix -> t -> bool
(** [check_suffix suff s] checks that string [s] ends with suffix [suff].
 Returns [true] when [suff] is empty. *)
;;

val chop_prefix : prefix -> t -> t
(** [chop_prefix pref s] returns a new string that contains a copy of string
[s] without prefix [pref].
 Returns [s] itself, not a copy of it, if [s] does not start with [pref]. *)
;;

val chop_suffix : suffix -> t -> t
(** [chop_suffix suff s] returns a new string that contains a copy of string
 [s] without suffix [suff].
 Returns [s] itself, not a copy of it, if [s] does not end with [suff]. *)
;;

val iso_latin1_to_us_ascii : (char -> t) -> t -> t
(** [iso_latin1_to_us_ascii of_unextended_char s] returns a new string that
 contains a copy of string [s] with all non US ASCII ISO Latin-1 (8859-1)
 characters translated into US ASCII characters.
 Digrams are replaced by the two corresponding characters and non alphabetic
 non US ASCII characters are treated by the function [of_unextended_char].

 Raise [Invalid_argument] if the result would be longer than
 [Sys.max_string_length]. *)
;;

val compare : t -> t -> int
(** The comparison function for strings, with the same specification as
 [Pervasives.compare].
 Along with type [t], this function [compare] allows module {!Ustring} to be
 passed as argument to functors [Set.Make] and [Map.Make]. *)
;;

val equal : t -> t -> bool
(** The equal function for strings. *)

(**/**)
(** {6 Undocumented features} *)

val unsafe_get : t -> index -> char;;
