(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: usubarray.ml,v 1.16 2017/10/25 13:01:01 fclement Exp $ *)

(* Operations on subarrays. *)

module Nat = Unatural;;
module L = Ulist;;
module A = Uarray;;
module Sat = Usubarray_types;;

let _0_ = Nat._0_;;

type length = Sat.length
and index = Sat.index
and offset = Sat.offset
and range = Sat.range =
  {
   offset : offset;
   length : length;
  }
and indices = Sat.indices =
  | Range of range
  | Indices of index array
and 'a t = 'a Sat.t = private
  {
   support : 'a array;
   indices : indices;
  }
and 'a subarray = 'a t
;;

let range = Sat.range
and indices = Sat.indices
;;

let fold_left_i f accu = function
  | { support = a; indices = Range { offset = ofs; length = len; }; } ->
      let rec loop accu i idx =
        (* i is increasing from 0, and idx = ofs + i. *)
        if i >= len then accu else
        (* Now, we have i < len, thus
           0 <= idx = ofs + i < ofs + len <= length a (see Sat.range).
           Therefore, the following A.unsafe_get will be safe. *)
        let x = A.unsafe_get a idx in
        let accu = f accu i idx x in
        loop accu (Nat.succ i) (Nat.succ idx) in
      loop accu _0_ ofs
  | { support = a; indices = Indices idxa; } ->
      A.fold_left_i
        (fun accu i idx ->
          (* Now, we have 0 <= i < length idxa, thus
             0 <= idx = idxa.(i) < length a (see Sat.indices).
             Therefore, the following A.unsafe_get will be safe. *)
          let x = A.unsafe_get a idx in
          f accu i idx x)
        accu idxa
;;

let fold_left f = fold_left_i (fun accu _i _idx x -> f accu x);;

let find p = function
  | { support = a; indices = Range { offset = ofs; length = len; }; } ->
      let rec loop i idx =
        (* i is increasing from 0, and idx = ofs + i. *)
        if i >= len then raise Not_found else
        (* Now, we have 0 <= i < len, thus
           0 <= idx = ofs + i < ofs + len <= length a (see Sat.range).
           Therefore, the following A.unsafe_get will be safe. *)
        let x = A.unsafe_get a idx in
        if p x then x else loop (Nat.succ i) (Nat.succ idx) in
      loop _0_ ofs
  | { support = a; indices = Indices idxa; } ->
      (* In the following, we have idx = idxa.(i) with 0 <= i < length idxa,
         thus 0 <= idx < length a (see Sat.indices).
         Therefore, both A.unsafe_get will be safe. *)
      let idx = A.find (fun idx -> p (A.unsafe_get a idx)) idxa in
      A.unsafe_get a idx
;;

let find_all p = function
  | { support = a; indices = Range { offset = ofs; length = len; }; } ->
      let rec loop accu i idx =
        (* i is increasing from 0, and idx = ofs + i. *)
        if i >= len then accu else
        (* Now, we have 0 <= i < len, thus
           0 <= idx = ofs + i < ofs + len <= length a (see Sat.range).
           Therefore, the following A.unsafe_get will be safe. *)
        let x = A.unsafe_get a idx in
        let accu = if p x then x :: accu else accu in
        loop accu (Nat.succ i) (Nat.succ idx) in
      L.rev (loop [] _0_ ofs)
  | { support = a; indices = Indices idxa; } ->
      (* In the following, we have idx = idxa.(i) with 0 <= i < length idxa,
         thus 0 <= idx < length a (see Sat.indices).
         Therefore, both A.unsafe_get will be safe. *)
      let idxs = A.rev_find_all (fun idx -> p (A.unsafe_get a idx)) idxa in
      L.map (A.unsafe_get a) idxs
;;
