(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uuniverse.mli,v 1.14 2017/05/31 15:24:06 fclement Exp $*)

(** {3 Universes}

 A universe is the nonempty support of a family of sets.
 It is a set that contains all the elements that belong to a set of the
 family of sets.
 In short, any set we consider in this library is a subset of some
 universe. *)

type 'a index = 'a Uindex.t;;

type 'a indices = 'a Uuniverse_types.indices;;
type 'a elements = 'a array;;

type 'a t = 'a Uuniverse_types.t = private
  {
   indices : 'a indices;
   mutable elements : 'a elements;
   mutable cardinal : Unatural.t;
  }
and 'a universe = 'a t
(** The type of universes. *)
;;

(** {6 Expansion phase, or transient state} *)

val make : Unatural.t -> 'a -> 'a t;;

val singleton : 'a -> 'a t;;

val add : 'a t -> 'a -> unit;;

val copy : 'a t -> 'a t;;

val product : 'a t -> 'b t -> ('a * 'b) t;;

val of_list1 : 'a Ulist1.t -> 'a t
(** [ of_list1 l1] creates the universe made of all distinct elements of the
 nonempty list [l1]. *)
;;

val of_array : 'a array -> 'a t
(** [of_array a] creates the universe made of all distinct elements of the
 array [a].

 Raise [Invalid_arg] if [a] is the empty array. *)
;;

val of_range : int -> int -> int t
(** [of_range i j] creates the universe for integers in the range
 [\[ i, j \[].

 Raise [Invalid_arg] if [i >= j]. *)
;;

val of_nat_range : Unatural.t -> Unatural.t -> Unatural.t t
(** Same as [of_range], but for natural numbers. *)
;;

val of_rectangle : int -> int -> (int * int) t
(** [of_rectangle h w] is equivalent to
 [product (of_range 0 h) (of_range 0 w)].

 Raise [Invalid_arg] if [h <= 0] or [w <= 0]. *)
;;

val of_nat_rectangle :
  Unatural.t -> Unatural.t -> (Unatural.t * Unatural.t) t
(** Same as [of_rectangle], but for natural numbers. *)
;;

(** {6 Steady state} *)

val stats : 'a t -> Hashtbl.statistics;;

val cardinal : 'a t -> Unatural.t;;

val valid_index : 'a t -> 'a index -> bool;;
val check_index : Upervasives.function_name -> 'a t -> 'a index -> unit;;

val index : 'a t -> 'a -> 'a index;;
val indices : 'a t -> 'a index t;;

val element : 'a t -> 'a index -> 'a;;
val elements : 'a t -> 'a array;;
val sub_elements : 'a t -> 'a index array -> 'a array;;

val mem : 'a t -> 'a -> bool;;

val fold_left : ('accu -> 'a -> 'accu) -> 'accu -> 'a t -> 'accu;;
val fold_left_i :
  ('accu -> 'a index -> 'a -> 'accu) -> 'accu -> 'a t -> 'accu
;;

val iter : ('a -> unit) -> 'a t -> unit;;
val iter_i : ('a index -> 'a -> unit) -> 'a t -> unit;;

val map_elements : ('a -> 'b) -> 'a t -> 'b array;;

val map : ('a -> 'b) -> 'a t -> 'b t;;

val find : ('a -> bool) -> 'a t -> 'a;;

val find_all : ('a -> bool) -> 'a t -> 'a list;;

val to_list : 'a t -> 'a list;;
