(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uunix.ml,v 1.46 2017/05/31 15:24:06 fclement Exp $ *)

module U = Unix;;

module Nat = Unatural;;
module S = Ustring;;
module A = Uarray;;

let _0_ = Nat._0_;;

let a_concat_string sa = S.concat " " (A.to_list sa);;

type program_name = Ufilename.filename
and program_arg = string
and program_args = program_arg list
and program_argv = program_arg array
and binding = string
and program_environment = binding array
;;

let qual = Upervasives.qualify "Uunix";;
let failwith = Upervasives.failwith_qual qual;;

let check_program_is_first_argument fun_name prog args =
  if A.length args = _0_ ||
    (* Now, we have A.length > 0.
       Therefore, the following A.unsafe_get will be safe. *)
    A.unsafe_get args _0_ <> prog then
  let cl = a_concat_string args in
  failwith fun_name
    (Printf.sprintf
       "progam %S is not the first argument of the argument list %S"
       prog cl)
;;

let gen_execv exec_name exec prog args =
  check_program_is_first_argument exec_name prog args;
  U.handle_unix_error exec prog args
;;

let execv = gen_execv "execv" U.execv
and execvp = gen_execv "execvp" U.execvp
and execve = gen_execv "execve" U.execve
and execvpe = gen_execv "execvpe" U.execvpe
;;

type process_io_channels = Scanf.Scanning.in_channel * Pervasives.out_channel;;

let spawn_init init_fun prog io_ref =
  match !io_ref with
  | Some io -> io
  | None ->
    let (ic, oc) = U.handle_unix_error U.open_process prog in
    let ib = Scanf.Scanning.from_channel ic in
    let io = (ib, oc) in
    io_ref := Some io;
    init_fun io;
    io
;;

let spawn = spawn_init ignore;;

type process_io_buffers =
  {
   input_buffer : Scanf.Scanning.scanbuf;
   output_buffer : Buffer.t;
   flush_output : unit -> unit;
   close_io_channels : unit -> unit;
  }
;;

let pio_spawn spf prog piob_ref =
  match !piob_ref with
  | Some piob -> piob
  | None ->
    let (ib, oc) = spawn prog (ref None) in
    spf.Udebug.spf_report (Printf.sprintf "program %S spawned" prog);
    let ob = Buffer.create 1024 in
    let flush () = Ubuffer.flush_buffer oc ob in
    let close () =
      Scanf.Scanning.close_in ib;
      Pervasives.close_out oc; in
    let piob =
      { input_buffer = ib; output_buffer = ob;
        flush_output = flush; close_io_channels = close; } in
    piob_ref := Some piob;
    piob
;;

let std_process_io_buffers =
  let ib = Scanf.Scanning.stdin in
  let ob = Buffer.create 1024 in
  let flush () = Ubuffer.flush_buffer_on_stdout ob in
  let close = ignore in
  { input_buffer = ib; output_buffer = ob;
    flush_output = flush; close_io_channels = close; }
;;

(* Connected processes stuff. *)

type pid = int;;

let connect_stdio proc (fdin1, fdout1) (fdin2, fdout2) =
  let pid = U.fork () in
  match pid with
  | 0 ->
    U.close fdout1;
    U.close fdin2;
    U.dup2 fdin1 U.stdin;
    U.close fdin1;
    U.dup2 fdout2 U.stdout;
    U.close fdout2;
    proc ()
  | pid -> pid
;;

let connect_bi_directional proc1 proc2 =
  let p1 = U.pipe () in
  let p2 = U.pipe () in
  let pid1 = connect_stdio proc1 p2 p1 in
  let pid2 = connect_stdio proc2 p1 p2 in
  (pid1, pid2)
;;

let launch prog args () = execv prog (A.of_list (prog :: args));;

let launch_connected_processes prog1 args1 prog2 args2 =
  U.handle_unix_error
    connect_bi_directional (launch prog1 args1) (launch prog2 args2)
;;

let connect_stdio_with_err proc (fdin1, fdout1) (fdin2, fdout2) fderr =
  let pid = U.fork () in
  match pid with
  | 0 ->
    U.close fdout1;
    U.close fdin2;
    U.dup2 fdin1 U.stdin;
    U.close fdin1;
    U.dup2 fdout2 U.stdout;
    U.close fdout2;
    U.dup2 fderr U.stderr;
    U.close fderr;
    proc ()
  | pid -> pid
;;

let connect_bi_directional_with_err proc1 fderr1 proc2 fderr2 =
  let p1 = U.pipe () in
  let p2 = U.pipe () in
  let pid1 = connect_stdio_with_err proc1 p2 p1 fderr1 in
  let pid2 = connect_stdio_with_err proc2 p1 p2 fderr2 in
  (pid1, pid2)
;;

let launch_connected_processes_with_err
    prog1 args1 fderr1 prog2 args2 fderr2 =
  connect_bi_directional_with_err
    (launch prog1 args1) fderr1
    (launch prog2 args2) fderr2
;;

let launch_connected_processes_with_err_on_channels
    prog1 args1 oc1 prog2 args2 oc2 =
  let fderr1 = U.handle_unix_error U.descr_of_out_channel oc1 in
  let fderr2 = U.handle_unix_error U.descr_of_out_channel oc2 in
  U.handle_unix_error
    launch_connected_processes_with_err
    prog1 args1 fderr1 prog2 args2 fderr2
;;

let wait_for_normal_termination pid =
  let (_pid, status) = U.handle_unix_error U.waitpid [] pid in
  let error pid msg n ier =
    prerr_endline (Printf.sprintf "Process %d %s %d." pid msg n);
    exit ier in
  match status with
  | U.WEXITED 0 -> ()
  | U.WEXITED n -> error pid "exited with error" n n;
  | U.WSIGNALED n -> error pid "was killed by signal" n 1;
  | U.WSTOPPED n -> error pid "was stopped by signal" n 1;
;;

let check_termination pid =
  ignore (U.handle_unix_error U.waitpid [] pid)
;;

let kill_and_wait pid =
  U.handle_unix_error U.kill pid Sys.sigterm;
  check_termination pid;
;;

let is_dead pid =
  let (pid, _) = U.handle_unix_error U.waitpid [ U.WNOHANG; ] pid in
  pid > 0
;;

let is_one_dead = Ulist.exists is_dead;;
